@@ -3,7 +3,7 @@
  * logtofile.c
  *      pgaudit addon to redirect audit log lines to an independent file
  *
- * Copyright (c) 2020-2021, Francisco Miguel Biete Banon
+ * Copyright (c) 2020-2023, Francisco Miguel Biete Banon
  * Copyright (c) 2014, 2ndQuadrant Ltd.
  *
  * This code is released under the PostgreSQL licence, as given at
@@ -34,14 +34,6 @@
 /* Defines */
 #define PGAUDIT_PREFIX_LINE "AUDIT: "
 #define PGAUDIT_PREFIX_LINE_LENGTH sizeof(PGAUDIT_PREFIX_LINE) - 1
-#define INTERCEPT_DISCONNECTION_PREFIX "disconnection: session time:"
-#define INTERCEPT_DISCONNECTION_PREFIX_LENGTH sizeof(INTERCEPT_DISCONNECTION_PREFIX) - 1
-#define INTERCEPT_CONNECTION_PREFIX1 "connection authorized: user="
-#define INTERCEPT_CONNECTION_PREFIX1_LENGTH sizeof(INTERCEPT_CONNECTION_PREFIX1) - 1
-#define INTERCEPT_CONNECTION_PREFIX2 "connection received: host="
-#define INTERCEPT_CONNECTION_PREFIX2_LENGTH sizeof(INTERCEPT_CONNECTION_PREFIX2) - 1
-#define INTERCEPT_CONNECTION_PREFIX3 "password authentication failed for user"
-#define INTERCEPT_CONNECTION_PREFIX3_LENGTH sizeof(INTERCEPT_CONNECTION_PREFIX3) - 1
 #define FORMATTED_TS_LEN 128
 
 /*
@@ -55,18 +47,62 @@
 #define LBF_MODE _IOLBF
 #endif
 
+
+/* Extracted from src/backend/po */
+static const char * postgresConnMsg[] = {
+  "connection received: host=%s port=%s",
+  "connection received: host=%s",
+  "connection authorized: user=%s",
+  "connection authenticated: identity=\"%s\" method=%s (%s:%d)",
+  "replication connection authorized: user=%s",
+  "replication connection authorized: user=%s SSL enabled (protocol=%s, cipher=%s, bits=%d, compression=%s)",
+  "replication connection authorized: user=%s application_name=%s",
+  "replication connection authorized: user=%s application_name=%s SSL enabled (protocol=%s, cipher=%s, bits=%d, compression=%s)",
+  "password authentication failed for user \"%s\"",
+  "authentication failed for user \"%s\": host rejected",
+  "\"trust\" authentication failed for user \"%s\"",
+  "Ident authentication failed for user \"%s\"",
+  "Peer authentication failed for user \"%s\"",
+  "password authentication failed for user \"%s\"",
+  "SSPI authentication failed for user \"%s\"",
+  "PAM authentication failed for user \"%s\"",
+  "BSD authentication failed for user \"%s\"",
+  "LDAP authentication failed for user \"%s\"",
+  "certificate authentication failed for user \"%s\"",
+  "RADIUS authentication failed for user \"%s\"",
+  "authentication failed for user \"%s\": invalid authentication method",
+  "connection authorized: user=%s database=%s",
+  "connection authorized: user=%s database=%s SSL enabled (protocol=%s, cipher=%s, bits=%d, compression=%s)",
+  "connection authorized: user=%s database=%s application_name=%s",
+  "connection authorized: user=%s database=%s application_name=%s SSL enabled (protocol=%s, cipher=%s, bits=%d, compression=%s)",
+};
+
+/* Extracted from src/backend/po */
+static const char * postgresDisconnMsg[] = {
+  "disconnection: session time: %d:%02d:%02d.%03d user=%s database=%s host=%s%s%s"
+};
+
 /* Buffers for formatted timestamps */
 static char formatted_start_time[FORMATTED_TS_LEN];
 static char formatted_log_time[FORMATTED_TS_LEN];
 
 /* SHM structure */
+typedef struct pgAuditLogToFilePrefix {
+  char *prefix;
+  int length;
+} pgAuditLogToFilePrefix;
+
 typedef struct pgAuditLogToFileShm {
   LWLock *lock;
-
   bool force_rotation;
+  pgAuditLogToFilePrefix **prefixes_connection;
+  size_t num_prefixes_connection;
+  pgAuditLogToFilePrefix **prefixes_disconnection;
+  size_t num_prefixes_disconnection;
 } pgAuditLogToFileShm;
 
 static pgAuditLogToFileShm *pgaudit_log_shm = NULL;
+static bool pgAuditLogToFileShutdown = false;
 
 /* Audit log file handler */
 static FILE *file_handler = NULL;
@@ -85,12 +121,20 @@ bool guc_pgaudit_log_disconnections = false;
 /* Old hook storage for loading/unloading of the extension */
 static emit_log_hook_type prev_emit_log_hook = NULL;
 static shmem_startup_hook_type prev_shmem_startup_hook = NULL;
+#if (PG_VERSION_NUM >= 150000)
+static shmem_request_hook_type prev_shmem_request_hook = NULL;
+#endif
 
 /* Hook functions */
 static void pgauditlogtofile_emit_log(ErrorData *edata);
 static void pgauditlogtofile_shmem_startup(void);
+#if (PG_VERSION_NUM >= 150000)
+static void pgauditlogtofile_shmem_request(void);
+#endif
 
 /* Internal functions */
+static char ** pgauditlogtofile_unique_prefixes(const char **messages, const size_t num_messages, size_t *num_unique);
+
 static void guc_assign_directory(const char *newval, void *extra);
 static void guc_assign_filename(const char *newval, void *extra);
 static bool guc_check_directory(char **newval, void **extra, GucSource source);
@@ -105,9 +149,11 @@ static void pgauditlogtofile_format_log_time(void);
 static void pgauditlogtofile_format_start_time(void);
 static bool pgauditlogtofile_is_enabled(void);
 static bool pgauditlogtofile_is_open_file(void);
+static bool pgauditlogtofile_is_prefixed(const char *msg);
 static bool pgauditlogtofile_needs_rotate_file(void);
 static bool pgauditlogtofile_open_file(void);
 static bool pgauditlogtofile_record_audit(const ErrorData *edata, int exclude_nchars);
+static void pgauditlogtofile_shmem_shutdown(int code, Datum arg);
 static bool pgauditlogtofile_write_audit(const ErrorData *edata, int exclude_nchars);
 
 
@@ -195,8 +241,13 @@ void _PG_init(void) {
 
   EmitWarningsOnPlaceholders("pgauditlogtofile");
 
+#if (PG_VERSION_NUM >= 150000)
+	prev_shmem_request_hook = shmem_request_hook;
+	shmem_request_hook = pgauditlogtofile_shmem_request;
+#else
   RequestAddinShmemSpace(MAXALIGN(sizeof(pgAuditLogToFileShm)));
   RequestNamedLWLockTranche("pgauditlogtofile", 1);
+#endif
 
   prev_shmem_startup_hook = shmem_startup_hook;
   shmem_startup_hook = pgauditlogtofile_shmem_startup;
@@ -212,11 +263,23 @@ void _PG_fini(void) {
   shmem_startup_hook = prev_shmem_startup_hook;
 }
 
+#if (PG_VERSION_NUM >= 150000)
+static void pgauditlogtofile_shmem_request(void) {
+  if (prev_shmem_request_hook)
+		prev_shmem_request_hook();
+
+  RequestAddinShmemSpace(MAXALIGN(sizeof(pgAuditLogToFileShm)));
+  RequestNamedLWLockTranche("pgauditlogtofile", 1);
+}
+#endif
+
 /*
  * SHMEM startup hook - Initialize SHMEM structure
  */
 static void pgauditlogtofile_shmem_startup(void) {
   bool found;
+  size_t num_messages, i, j;
+  char **prefixes;
 
   if (prev_shmem_startup_hook)
     prev_shmem_startup_hook();
@@ -227,18 +290,100 @@ static void pgauditlogtofile_shmem_startup(void) {
   LWLockAcquire(AddinShmemInitLock, LW_EXCLUSIVE);
   pgaudit_log_shm = ShmemInitStruct("pgauditlogtofile", sizeof(pgAuditLogToFileShm), &found);
   if (!found) {
+    // Get unique prefixes and copy them to SHMEM
+    num_messages = sizeof(postgresConnMsg) / sizeof(char *);
+    prefixes = pgauditlogtofile_unique_prefixes(postgresConnMsg, num_messages, &pgaudit_log_shm->num_prefixes_connection);
+    pgaudit_log_shm->prefixes_connection = ShmemAlloc(pgaudit_log_shm->num_prefixes_connection * sizeof(pgAuditLogToFilePrefix *));
+    for (i = 0, j = 0; i < num_messages; i++) {
+      if (prefixes[i] != NULL) {
+        pgaudit_log_shm->prefixes_connection[j] = ShmemAlloc(sizeof(pgAuditLogToFilePrefix));
+        pgaudit_log_shm->prefixes_connection[j]->length = strlen(prefixes[i]);
+        pgaudit_log_shm->prefixes_connection[j]->prefix = ShmemAlloc( (pgaudit_log_shm->prefixes_connection[j]->length + 1) * sizeof(char) );
+        strcpy(pgaudit_log_shm->prefixes_connection[j]->prefix, prefixes[i]);
+        free(prefixes[i]);
+        j++;
+      }
+    }
+    free(prefixes);
+
+    num_messages = sizeof(postgresDisconnMsg) / sizeof(char *);
+    prefixes = pgauditlogtofile_unique_prefixes(postgresDisconnMsg, num_messages, &pgaudit_log_shm->num_prefixes_disconnection);
+    pgaudit_log_shm->prefixes_disconnection = ShmemAlloc(pgaudit_log_shm->num_prefixes_disconnection * sizeof(pgAuditLogToFilePrefix *));
+    for (i = 0, j = 0; i < num_messages; i++) {
+      if (prefixes[i] != NULL) {
+        pgaudit_log_shm->prefixes_disconnection[j] = ShmemAlloc(sizeof(pgAuditLogToFilePrefix));
+        pgaudit_log_shm->prefixes_disconnection[j]->length = strlen(prefixes[i]);
+        pgaudit_log_shm->prefixes_disconnection[j]->prefix = ShmemAlloc( (pgaudit_log_shm->prefixes_disconnection[j]->length + 1) * sizeof(char) );
+        strcpy(pgaudit_log_shm->prefixes_disconnection[j]->prefix, prefixes[i]);
+        free(prefixes[i]);
+        j++;
+      }
+    }
+    free(prefixes);
+
     pgaudit_log_shm->lock = &(GetNamedLWLockTranche("pgauditlogtofile"))->lock;
     pgaudit_log_shm->force_rotation = false;
-    pgauditlogtofile_calculate_next_rotation_time();
+    if (guc_pgaudit_log_rotation_age > 0)
+      pgauditlogtofile_calculate_next_rotation_time();
     pgauditlogtofile_calculate_filename();
   }
   LWLockRelease(AddinShmemInitLock);
 
+  if (!IsUnderPostmaster)
+    on_shmem_exit(pgauditlogtofile_shmem_shutdown, (Datum) 0);
+
   if (!found) {
     ereport(LOG, (errmsg("pgauditlogtofile extension initialized")));
   }
 }
 
+static char ** pgauditlogtofile_unique_prefixes(const char **messages, const size_t num_messages, size_t *num_unique) {
+  bool is_unique;
+  char **prefixes;
+  char *message, *prefix, *dup;
+  size_t i, j;
+
+  *num_unique = 0;
+
+  prefixes = malloc(num_messages * sizeof(char *));
+
+  for (i = 0; i < num_messages; i++) {
+#ifdef ENABLE_NLS
+    // Get translation - static copy
+    message = gettext(messages[i]);
+#else
+    // Pointer to original = static copy
+    message = messages[i];
+#endif
+    // Get a copy that we can modify
+    dup = strdup(message);
+    prefix = strtok(dup, "%");
+    if (prefix != NULL) {
+      // Search duplicated
+      is_unique = true;
+      for (j = 0; j < i; j++) {
+        if (prefixes[j] != NULL) {
+          if (strcmp(prefixes[j], prefix) == 0) {
+            // Skip - prefix already present
+            is_unique = false;
+          }
+        }
+      }
+
+      if (is_unique) {
+        prefixes[i] = malloc((strlen(prefix) + 1) * sizeof(char));
+        strcpy(prefixes[i], prefix);
+        *num_unique += 1;
+      } else {
+        prefixes[i] = NULL;
+      }
+    }
+    free(dup);
+  }
+
+  return prefixes;
+}
+
 /*
  * Hook to emit_log - write the record to the audit or send it to the default
  * logger
@@ -252,19 +397,7 @@ static void pgauditlogtofile_emit_log(ErrorData *edata) {
       exclude_nchars = PGAUDIT_PREFIX_LINE_LENGTH;
       edata->output_to_server = false;
     }
-    else if (pg_strncasecmp(edata->message, INTERCEPT_CONNECTION_PREFIX1, INTERCEPT_CONNECTION_PREFIX1_LENGTH) == 0) {
-      edata->output_to_server = false;
-      exclude_nchars = 0;
-    }
-    else if (pg_strncasecmp(edata->message, INTERCEPT_CONNECTION_PREFIX2, INTERCEPT_CONNECTION_PREFIX2_LENGTH) == 0) {
-      edata->output_to_server = false;
-      exclude_nchars = 0;
-    }
-    else if (pg_strncasecmp(edata->message, INTERCEPT_CONNECTION_PREFIX3, INTERCEPT_CONNECTION_PREFIX3_LENGTH) == 0) {
-      edata->output_to_server = false;
-      exclude_nchars = 0;
-    }
-    else if (pg_strncasecmp(edata->message, INTERCEPT_DISCONNECTION_PREFIX, INTERCEPT_DISCONNECTION_PREFIX_LENGTH) == 0) {
+    else if (pgauditlogtofile_is_prefixed(edata->message)) {
       edata->output_to_server = false;
       exclude_nchars = 0;
     }
@@ -285,10 +418,9 @@ static void pgauditlogtofile_emit_log(ErrorData *edata) {
  * Checks if pgauditlogtofile is completely started and configured
  */
 static inline bool pgauditlogtofile_is_enabled(void) {
-  if (!pgaudit_log_shm || guc_pgaudit_log_directory == NULL ||
-      guc_pgaudit_log_filename == NULL ||
-      strlen(guc_pgaudit_log_directory) == 0 ||
-      strlen(guc_pgaudit_log_filename) == 0)
+  if (pgAuditLogToFileShutdown || !pgaudit_log_shm ||
+      guc_pgaudit_log_directory == NULL || guc_pgaudit_log_filename == NULL ||
+      strlen(guc_pgaudit_log_directory) == 0 || strlen(guc_pgaudit_log_filename) == 0)
     return false;
 
   return true;
@@ -334,12 +466,34 @@ static inline bool pgauditlogtofile_is_open_file(void) {
     return false;
 }
 
+/*
+ * Checks if a message starts with one of our intercept prefixes
+ */
+static inline bool pgauditlogtofile_is_prefixed(const char *msg) {
+  bool found = false;
+  size_t i;
+
+  if (guc_pgaudit_log_connections) {
+    for (i = 0; !found && i < pgaudit_log_shm->num_prefixes_connection; i++) {
+      found = pg_strncasecmp(msg, pgaudit_log_shm->prefixes_connection[i]->prefix, pgaudit_log_shm->prefixes_connection[i]->length) == 0;
+    }
+  }
+
+  if (!found && guc_pgaudit_log_disconnections) {
+    for (i = 0; !found && i < pgaudit_log_shm->num_prefixes_disconnection; i++) {
+      found = pg_strncasecmp(msg, pgaudit_log_shm->prefixes_disconnection[i]->prefix, pgaudit_log_shm->prefixes_disconnection[i]->length) == 0;
+    }
+  }
+
+  return found;
+}
+
 /*
  * Checks if the audit log file needs to be rotated before we use it
  */
 static bool pgauditlogtofile_needs_rotate_file(void) {
   /* Rotate if we are forcing */
-  if (pgaudit_log_shm->force_rotation) {
+  if (pgaudit_log_shm->force_rotation && MyProc != NULL) {
     LWLockAcquire(pgaudit_log_shm->lock, LW_EXCLUSIVE);
     pgaudit_log_shm->force_rotation = false;
     LWLockRelease(pgaudit_log_shm->lock);
@@ -348,7 +502,7 @@ static bool pgauditlogtofile_needs_rotate_file(void) {
 
   /* Rotate if rotation_age is exceeded, and this backend is the first in notice
    * it */
-  if ((pg_time_t)time(NULL) >= next_rotation_time) {
+  if (guc_pgaudit_log_rotation_age > 0 && (pg_time_t)time(NULL) >= next_rotation_time) {
     pgauditlogtofile_calculate_next_rotation_time();
     return true;
   }
@@ -431,9 +585,15 @@ static bool pgauditlogtofile_open_file(void) {
  */
 static void pgauditlogtofile_calculate_filename(void) {
   int len;
-  pg_time_t current_rotation_time =
-      next_rotation_time -
-      guc_pgaudit_log_rotation_age * SECS_PER_MINUTE;
+  pg_time_t current_rotation_time;
+  if (guc_pgaudit_log_rotation_age > 0) {
+    current_rotation_time =
+         next_rotation_time -
+         guc_pgaudit_log_rotation_age * SECS_PER_MINUTE;
+  } else {
+    current_rotation_time = (pg_time_t)time(NULL);
+  }
+
 
   memset(filename, 0, sizeof(filename));
   snprintf(filename, MAXPGPATH, "%s/",
@@ -670,3 +830,10 @@ static inline void pgauditlogtofile_format_log_time(void) {
   sprintf(msbuf, ".%03d", (int)(tv.tv_usec / 1000));
   memcpy(formatted_log_time + 19, msbuf, 4);
 }
+
+/*
+ * Identify when we are doing a shutdown
+ */
+static void pgauditlogtofile_shmem_shutdown(int code, Datum arg) {
+  pgAuditLogToFileShutdown = true;
+}
